####################################################
# CopyRight    : (c) Riello UPS GmbH               #
# Author       : Ruediger Brink                    #
# Name         : HVshutdown.ps1                    #
# Version      : 1.00                              #
# Date         : 09/11/2012                        #
# Internet     : http://www.riello-ups.de          #
# Last modified: 18/10/2018                        #
#################################################### 

$global:ScriptName = $myInvocation.MyCommand.Name

$global:startTime  = Get-Date

#####################################################
# If $RUN_EXEC is $false then you can start the     #
# script ups_shut.bat to see what happens. No action#
# will be performed, only a log file is written.    #
#                                                   #
# Possible values: $true or $false                  #
#####################################################
#
$global:RUN_EXEC   = $false

#####################################################
# DON'T DELETE THIS ARRAY DECLARATION!!!!!!        #
#####################################################
#
$global:HV_SERVER  = @()

#####################################################
# Please add all your Hyper-V hosts here.           #
#####################################################
#
#$global:HV_SERVER += ,@("RPS-HYPER2012")
#$global:HV_SERVER += ,@("RPS-HYPER2008")
$global:HV_SERVER += ,@("RPS-HYPER2016")

#####################################################
# DON'T DELETE THIS DECLARATIONs                    #
#####################################################
#
$global:HV_HOST     = ""
$global:HV_OS       = ""
$global:HV_Major    = ""
$global:HV_Minor    = ""
$global:HV_Build    = ""
$global:HV_NSPACE   = ""
$global:HV_CAPTION  = ""
$global:HV_LANGUAGE = ""

#####################################################
# Where is the log file to be written.              #
# If this blank no log file will be written         #
#####################################################
#
$global:UPS_LOG    = "c:\ProgramData\Upsmon\ups.log"

#####################################################
# $DOWN_LOOP : counter of how often to be tested    #
#              whether VMs are shutdown             #
# $DOWN_SLEEP: delay in seconds for the next test   #
#####################################################
#
$global:DOWN_LOOP  = 3
$global:DOWN_SLEEP = 15

####################################################
# DON'T DELETE THIS ARRAY DECLARATION!!!!!!        #
####################################################
#
$global:VM_DEPENDS = @()

####################################################
# If you have dependencies between your VMs then   #
# you have to configure this part.                 #
# Fill array with values                           #
#                                                  #
# array[0] = Server name           [String]        #
# array[1] = Sequence of Shutdown  [Integer]       #
# array[2] = Delay in seconds      [Integer]       #
# array[3] = Host name             [String]        #
# e.g.                                             #
# $VM_DEPENDS += ,@("Server name", 0, 10, "HOST1") #
####################################################
#
#$global:VM_DEPENDS += ,@("Server 2012", 0, 10, "RPS-HYPER2012")
#$global:VM_DEPENDS += ,@("Windows 8", 1, 10, "RPS-HYPER2012")
#$global:VM_DEPENDS += ,@("SuSE 12.2", 2, 5, "RPS-HYPER2008")
#$global:VM_DEPENDS += ,@("Dummy", 3, 15, "RPS-HYPER2008")

###################################################
# Function : writeLog                             #
# Parameter: Text, CR (true/false)                #
# Used     : Get-Date, Out-File, echo             #
# Returns  :                                      #
###################################################
#
function writeLog {
  param([string]$LogText, [switch]$NewLine)

  if ( $LogText ) {
    if ( $NewLine ) {
      echo "" >>$UPS_LOG
    }

    $DATUM = Get-Date -uformat %d.%m.%Y" "%T
    $DATUM+" "+$LogText | Out-File $global:UPS_LOG -Append
  } else {
    echo "" >> $global:UPS_LOG
  }
}

####################################################
# Function : VM_Exists                             #
# Parameter: VM Name                               #
# Used     :                                       #
# Returns  : $true  if exists                      #
#          : $false if not exists                  #
####################################################
#
function VM_Exists ( $VM_NAME ) {
  $global:VM_ALL | foreach {
    if ( $_.ElementName -eq $VM_NAME ) {
      return $true
    }
  }

  return $false
}

####################################################
# Function : Test_Depend                           #
# Parameter: Name of VM to test                    #
# Used     : $VM_DEPENDS                           #
# Returns  : $true if $VM is in array $VM_DEPENDS  #
#          : $false if not                         #
####################################################
#
function Test_Depend( $VM ) {
  if ( $global:VM_DEPENDS.Count -eq 0 ) {
    return $false
  } else {
    $global:VM_DEPENDS | foreach {
      if ( ($_[3] -eq $global:HV_HOST) -and ($_[0] -eq $VM) ) {
        return $true
      }
    }

    return $false
  }
}

####################################################
# Function : Get_Depend                            #
# Parameter: array index of $VM_DEPENDS            #
# Used     : $VM_DEPENDS                           #
# Returns  : on found - located array entry        #
#          : otherwise - an empty array            #
####################################################
#
function Get_Depend( $NR ) {
  $global:VM_DEPENDS | foreach {
    if ( ($_[1] -eq $NR) ) {
      # -and (VM_Exists( $_[0] )) ) {
      return $_
    }
  }

  return @()
}

####################################################
# Function : Delete_Depend                         #
# Parameter: array of VMs                          #
# Used     : Test_Depend                           #
# Returns  : array of VMs without $VM_DEPENDS      #
####################################################
#
function Delete_Depend( $VMs ) {
  if ( $global:VM_DEPENDS.Count -gt 0 ) {
    for ( $i=0; $i -lt $global:VM_DEPENDS.Count; $i++ ) {
      if ( $global:VM_DEPENDS[$i][3] -eq $global:HV_HOST ) {
        break
      }
    }

    if ( $i -lt $global:VM_DEPENDS.Count ) {
      $Ret_VMs = @()
      $global:VM_DEPENDS | foreach {
        if ( $_[3] -eq $global:HV_HOST ) {

          $VMs | foreach {
            if ( !(Test_Depend( $_.ElementName )) ) {
              $Ret_VMs += $_
            }
          }
        }
      }

      return $Ret_VMs
    } else {
      return $VMs
    }
  } else {
    return $VMs
  }
}

###################################################
# Function : Get_VMs                              #
# Parameter:                                      #
# Used     :                                      #
# Returns  : pointer array of running VMs         #
###################################################
#
function Get_VMs() {
  $query   = "SELECT * FROM Msvm_ComputerSystem WHERE ProcessID>'0' and Caption='" + $global:HV_CAPTION + "'"
  #$query   = "SELECT * FROM Msvm_ComputerSystem WHERE Caption='" + $global:HV_CAPTION + "'"
  $Get_VMs = Get-WmiObject -Query $query -Namespace $global:HV_NSPACE -ComputerName $global:HV_HOST

  return $Get_VMs
}

###################################################
# Function : Get_VM                               #
# Parameter: Name of VM                           #
# Used     :                                      #
# Returns  : pointer to running VM                #
###################################################
#
function Get_VM( $VM ) {
  $query  = "SELECT * FROM Msvm_ComputerSystem WHERE ProcessID>'0' and ElementName='" + $VM + "'"
  $Get_VM = Get-WmiObject -Query $query -Namespace $global:HV_NSPACE -ComputerName $global:HV_HOST

  return $Get_VM
}

###################################################
# Function : Shut_VM                              #
# Parameter: pointer to VM                        #
# Used     :                                      #
# Returns  : result of shutdown command           #
###################################################
#
function Shut_VM( $VM ) {
  writeLog("Execute shutdown : " + $VM.ElementName)

  $query    = "SELECT * FROM Msvm_ShutdownComponent WHERE SystemName='" + $VM.Name + "'"
  $Shutdown = Get-WmiObject -Query $query -Namespace $global:HV_NSPACE -ComputerName $global:HV_HOST
  $ResultVM = $Shutdown.InitiateShutdown($true, "Power failure! UPS shutdown")

  #writeLog("Result shutdown  : $ResultVM")

  return $ResultVM
}

function Get_HVInfo( $IS_HOST ) {
  $global:HV_OS     = (Get-WmiObject -Class Win32_OperatingSystem -ComputerName $IS_HOST).Caption
  $OSVersion = ((Get-WmiObject -Class Win32_OperatingSystem -ComputerName $IS_HOST).Version).Split("{.}")

  $global:HV_Major  = $OSVersion[0]
  $global:HV_Minor  = $OSVersion[1]
  $global:HV_Build  = $OSVersion[2]

  if ($global:HV_Minor -eq 1 -and (($global:HV_Build -eq 7600) -or ($global:HV_Build -eq 7601))) {
    $global:HV_NSPACE = "root\virtualization"
  } else {
    $global:HV_NSPACE = "root\virtualization\v2"
  }

  $LCID = (Get-WmiObject -Class Win32_OperatingSystem -ComputerName $IS_HOST).OSLanguage
  if ($LCID -eq "1031") {
    $global:HV_CAPTION = "Virtueller Computer"
  } elseif ($LCID -eq "1033") {
    $global:HV_CAPTION = "Virtual Machine"
  } elseif ($LCID -eq "1040") {
    $global:HV_CAPTION = "Macchina virtuale"
  } elseif ($LCID -eq "3082") {
    $CAPTION = "Maquina Virtual"
  } else {
    echo "" >> $UPS_LOG
    "Unknown LCID: "+$LCID | Out-File $UPS_LOG -Append
    "Program aborted!" | Out-File $UPS_LOG -Append
    Exit 1
  }

  $global:HV_LANGUAGE = (Get-WmiObject -Class Win32_OperatingSystem -ComputerName $IS_HOST).MUILanguages[0]
}
 
###################
#                 #
#     MAIN        #
#                 #
###################
#
if ( $global:RUN_EXEC ) {
  $global:RUN_MODE = "REAL"
} else {
  $global:RUN_MODE = "TEST"
}

writeLog -LogText "Start             : $global:ScriptName [$global:RUN_MODE MODE]`n" -NewLine

$global:HV_SERVER | foreach {
  $global:HV_HOST = $($_[0])

  Get_HVInfo( $global:HV_HOST )

  writeLog("Hostname          : $global:HV_HOST`n")
  writeLog("Major             : $global:HV_Major")
  writeLog("Minor             : $global:HV_Minor")
  writeLog("Build             : $global:HV_Build")
  writeLog("OS Version        : $global:HV_OS")
  writeLog("OS Language       : $global:HV_LANGUAGE`n")

  $global:VM_ALL   = Get_VMs
  $global:VM_ALONE = Delete_Depend( $global:VM_ALL )

  ######################################################################
  # shutdown all VMs where PowerState is PoweredOn but without $UPS_VM #
  ######################################################################
  #
  if ( $global:VM_ALONE ) {
    writeLog(">>> Shutdown all independent VM(s)..")

    foreach ( $VM in $global:VM_ALONE ) {
      if ( $global:RUN_EXEC ) {
        $ResultVM = Shut_VM( $VM )
      } else {
        writeLog("Simulate shutdown : " + $VM.ElementName)
      }
    }

    writeLog("")
  } else {
    writeLog("no independent VM(s) found for Shutdown!`n");
  }

  ##########################################
  # Loop to wait for shutdown of all VMs   #
  # maximal $DOWN_LOOP x $DOWN_SLEEP times #
  ##########################################
  #
  if ( $global:VM_ALONE ) {
    if ( $global:RUN_EXEC ) {
      writeLog -LogText "Wait $global:DOWN_SLEEP seconds, to give VM(s) time for shutdown." -NewLine
      Start-Sleep -s $global:DOWN_SLEEP
      $global:DOWN_LOOP -= 1

      $global:VM_ALL   = Get_VMs
      $global:VM_ALONE = Delete_Depend( $global:VM_ALL )
      while (( $global:VM_ALONE ) -and ( $global:DOWN_LOOP -gt 0 )) {
        if ( $global:VM_ALONE ) {
          foreach ( $VM in $global:VM_ALONE ) {
            writeLog("Wait for shutdown : " + $VM.ElementName)
          }
        }
        writeLog("Wait $global:DOWN_SLEEP seconds more..")
        Start-Sleep -s $global:DOWN_SLEEP

        $global:DOWN_LOOP -= 1
        $global:VM_ALL   = Get_VMs
        $global:VM_ALONE = Delete_Depend( $global:VM_ALL )
      }
      echo "" >> $global:UPS_LOG
    } else {
      writeLog("Simulate wait/loop: $global:DOWN_LOOP times / $global:DOWN_SLEEP seconds`n")
    }
  }
}

if ( $global:VM_DEPENDS.Count -gt 0 ) {
    writeLog(">>> Shutdown all dependent VM(s)..")

    for ( $i=0; $i -lt $global:VM_DEPENDS.Count; $i++ ) {
	  $VM_Get = Get_Depend( $i )

      if ( $VM_Get ) {
        $global:HV_HOST = $($VM_Get[3])
        #writelog("HOST: $global:HV_HOST")
		Get_HVInfo( $global:HV_HOST )
        $VM = Get_VM( $($VM_Get[0]) )
        if ( $VM ) {
          if ( $global:RUN_EXEC ) {
            $ResultVM = Shut_VM( $VM )
            if ( $($VM_Get[2]) -gt 0 ) {
              writeLog("Wait             : $($VM_Get[2]) seconds")
              Start-Sleep -s $($VM_Get[2])
            }
          } else {
            writeLog("Simulate shutdown : $($VM_Get[0]) on $($VM_Get[3])")
            if ( $($VM_Get[2]) -gt 0 ) {
              writeLog("Simulate wait     : $($VM_Get[2]) seconds")
            }
          }
        } else {
          writeLog("Error             : VM `"$($VM_Get[0])`" on $($VM_Get[3]) not running!")
        }
      } else {
        writeLog("Error             : No VM found with sequence `"$i`"!")
      }
    }
  writeLog("")
}

writeLog(">>> Shutdown all Hosts without UPS-Host..")
$UPS_HOST = (Get-WmiObject -Class Win32_OperatingSystem).CSName
$global:HV_SERVER | foreach {
  $HV_HOST = $($_[0])

  if ( $HV_HOST -ne $UPS_HOST ) {
    if ($global:RUN_EXEC) {
      writeLog("Execute shutdown  : $HV_HOST")
      Stop-Computer -ComputerName $HV_HOST -Force
    } else {
      writeLog("Simulate shutdown : $HV_HOST")
    }
  }
}

$endTime = Get-Date

writeLog("Total elapsed time: $(( $endTime - $global:startTime ).TotalSeconds) seconds`n")
writeLog("End               : $global:ScriptName [$global:RUN_MODE MODE]")
